import os
import sys
import time
import struct
import glob
import pickle

from subprocess import call
from bluepy import btle
from bluepy.btle import Scanner, DefaultDelegate

BLOCK_SIZE = 18

GALAXY_SERVICE_UUID = btle.UUID('05040302-0102-0304-0506-070807060504')
FIRMWARE_VERSION_CHARACTERISTIC_UUID = btle.UUID("08070605-0403-0201-0102-030405060708")
NOTIFICATION_CHARACTERISTIC_UUID = btle.UUID('3e64b283-25ea-4049-ae20-70cc81901343')
FIRMWARE_MODE_CHARACTERISTIC_UUID = btle.UUID('08070605-0403-0201-0102-03040506070F')
FIRMWARE_DATA_CHARACTERISTIC_UUID = btle.UUID('08070605-0403-0201-0102-030405060710')

def getCharacteristic(service, characteristicUuid):
    result = service.getCharacteristics(characteristicUuid)
    if len(result) > 0:
        return result[0]

class FrameReader:
    def __init__(self, path):
        self.__file = open(path, 'rb')
        self.__frames = {}
        self.size = os.path.getsize(path)
        self.blockCount = (self.size - 1) / BLOCK_SIZE + 1
        self.__nextFrame = 0

    def getFrame(self, frameNumber):
        if frameNumber in self.__frames:
            return self.__frames[frameNumber]
        if frameNumber == self.__nextFrame and not self.isAfterLastFrame(frameNumber):
            return self.__getNextFrame()

    def isAfterLastFrame(self, frameNumber):
        return frameNumber >= self.blockCount

    def __getNextFrame(self):
        data = self.__file.read(BLOCK_SIZE)
        frame = struct.pack('>H', self.__nextFrame) + data
        self.__frames[self.__nextFrame] = frame
        self.__nextFrame += 1
        return frame

class FirmwareUpdateDelegate(DefaultDelegate):
    def __init__(self, spu):
        DefaultDelegate.__init__(self)
        self.spu = spu

    def handleNotification(self, cHandle, data):
        update_code = data[0]
        frameNumber = struct.unpack('>H', data[1:])[0]
        if update_code == '\x01':
            self.spu.confirmFrame(frameNumber)
        elif update_code == '\x00':
            print 'Error, resending frame', frameNumber
            self.spu.resendFromFrame(frameNumber)
        elif update_code == '\x02':
            self.spu.finishUpdate()

class Spu:
    def __init__(self, dev):
        self.peripheral = None
        self.__dev = dev
        self.resent = False

    def getName(self):
        return self.__dev.getValueText(0x09)

    def getAddress(self):
        return self.__dev.addr

    def getRssi(self):
        return self.__dev.rssi

    def getValTxt(self, val):
        return self.__dev.getValueText(val)
        
    def getMe(self):
        return self.__dev
        
    def update(self, path):
        self.__initiateUpdate(path)
        self.__transfer() if self.__notificationCharacteristic else self.__transferForOldFirmware()
        print 'Update finished in', str(time.time() - self.__startTime) + 's'

    def finishUpdate(self):
        sys.stdout.write('\b\b\b100%\r\n')
        self.__updateFinished = True

    def connect(self):
        if self.peripheral:
            return
        print 'Connecting to %s (%s)' % (self.getName(), self.__dev.addr)
        self.peripheral = btle.Peripheral(self.__dev)
        self.peripheral.setDelegate(FirmwareUpdateDelegate(self))
        self.__retrieveCharacteristics()
        print 'Connected'

    def disconnect(self):
        self.peripheral.disconnect()
        self.peripheral = None

    def getFirmwareVersion(self):
        return str(self.__firmwareVersionCharacteristic.read())

    def __initiateUpdate(self, path):
        self.__startTime = time.time()
        self.__frameReader = FrameReader(path)
        self.__lastConfirmed = -1
        self.__updateFinished = False
        self.__nextFrame = 0
        self.__lastNotification = time.time()

        if self.__notificationCharacteristic:
            self.peripheral.writeCharacteristic(self.__notificationCharacteristic.getDescriptors()[0].handle, '\x01\x00', withResponse=True)
        start_order = '\x00' + struct.pack('>I', self.__frameReader.size)
        self.__firmwareModeCharacteristic.write(start_order, withResponse=True)
        call(["sudo", "hcitool", "lecup", "--handle", "64", "--min", "6", "--max", "7", "--latency", "0", "--timeout", "500"])
        time.sleep(2)

    def __transfer(self):
        sys.stdout.write('Updating: 00%')
        while not self.__updateFinished:
            self.__sendNextFrame()
            if time.time() - self.__lastNotification > 0.5:
                self.peripheral.waitForNotifications(0.001)
                self.__lastNotification = time.time()

    def __transferForOldFirmware(self):
        print 'Updating'
        while not self.__frameReader.isAfterLastFrame(self.__nextFrame):
            self.__sendNextFrame()
            time.sleep(0.002)
        self.__firmwareModeCharacteristic.write('\x01', withResponse=True)

    def confirmFrame(self, confirmedFrame):
        #sys.stdout.write('\b\b\b%02d%%' % (confirmedFrame * 100 / self.__frameReader.blockCount))
        sys.stdout.write('%02d%%' % (confirmedFrame * 100 / self.__frameReader.blockCount))
        self.__lastConfirmed = confirmedFrame
        self.__lastConfirmation = time.time()

    def resendFromFrame(self, nextFrame):
        self.__nextFrame = nextFrame - 1

    def __retrieveCharacteristics(self):
        service = self.peripheral.getServiceByUUID(GALAXY_SERVICE_UUID)
        self.__notificationCharacteristic = getCharacteristic(service, NOTIFICATION_CHARACTERISTIC_UUID)
        self.__firmwareModeCharacteristic = getCharacteristic(service, FIRMWARE_MODE_CHARACTERISTIC_UUID)
        self.__firmwareDataCharacteristic = getCharacteristic(service, FIRMWARE_DATA_CHARACTERISTIC_UUID)
        self.__firmwareVersionCharacteristic = getCharacteristic(service, FIRMWARE_VERSION_CHARACTERISTIC_UUID)

    def __sendNextFrame(self):
        frame = self.__frameReader.getFrame(self.__nextFrame)
        if frame:
            self.__firmwareDataCharacteristic.write(frame)
            self.__nextFrame += 1

def isSpu(dev):
    return dev.getValueText(0x07) == str(GALAXY_SERVICE_UUID)

def scanSpus():
    print 'Scanning SPUs'
    scanner = Scanner()
    devices = scanner.scan(3)
    return [Spu(dev) for dev in devices if isSpu(dev)]

def displayChoices(spus):
    for i, spu in enumerate(spus):
        print "%d. %s %s RSSI=%d" % (i + 1, spu.getName(), spu.getAddress(), spu.getRssi())

def selectSpus(spus, selected):
    userInput = raw_input("SPUs to update (separate with spaces, leave empty for all):\r\n")
    if userInput == '':
        return spus
    spuIndexes = map(lambda i: int(i) - 1, userInput.split(' '))
    return [spus[i] for i in spuIndexes if i in range(len(spus))]

def launchUpdater(path, updateAll):
    spus = scanSpus()
    if updateAll:
        spusToUpdate = spus
    else:
        displayChoices(spus)
        spusToUpdate = selectSpus(spus,'')
    for spu in spusToUpdate:
        updateSpu(spu, path, updateAll)

def updateSpu(spu, path, force):
    spu.connect()
    confirmUpdate = '' if force else raw_input("%s - firmware version: %s, update? (Y/n) " % (spu.getName(), spu.getFirmwareVersion()))
    if confirmUpdate == '' or confirmUpdate == 'y' or confirmUpdate == 'Y':
        spu.update(path)
    spu.disconnect()

def printHelp():
    print "usage: sudo python", sys.argv[0], "[-y] [<path>] [help]"
    print "    -y      update all SPUs (does not prompt the user)"
    print "    <path>  path to the kbin ('lg-emergency_.kbin' by default)"
    print "    help    print this help"

# def's below here are added by me
def getFirmware():
    for file in glob.glob("firmware/*.kbin"):
        # print(file)
        return file
        
def saveSpus(spus, dict):
    # config_spus = {}
    with open(dict, 'wb') as config_spus_file:
        pickle.dump(spus, config_spus_file)
        
def loadSpus(dict):
    with open(dict, 'rb') as config_dictionary_file:
        return pickle.load(config_dictionary_file)
        

if __name__ == "__main__":
    if "help" in sys.argv[1:]:
        printHelp()
    else:
        # list the current firmware in the firmware dir        
        path = ""
        # getFirmware()
        # print "Current path is ", path
        updateAll = False
        # print('in python')
        # print(sys.argv[1])
        for arg in sys.argv[1:]:
            #print(arg)
            if arg == '-y':
                updateAll = True
                launchUpdater(path, updateAll)
            elif arg == 'scan':
                # do something here to scan for devices
                spus = scanSpus()
                displayChoices(spus)
                saveSpus(spus, 'spu.dictionary')
            elif arg == 'update':
                # print ('in update')
                spus = loadSpus('spu.dictionary')
                path = sys.argv[2]  # firmware file and index of spu to upload
                num = int(sys.argv[3])
                # time.sleep(2)
                # print(path, num)
                updateSpu(spus[num-1], path, 'y')                                               
            else:
                path = arg
                # print ('in else')
                #launchUpdater(path, updateAll)
